# -*- coding: utf-8 -*-

"""
gspread_dataframe
~~~~~~~~~~~~~~~~~

This module contains functions to retrieve a gspread worksheet as a
`pandas.DataFrame`, and to set the contents of a worksheet
using a `pandas.DataFrame`. To use these functions, have
Pandas 0.14.0 or greater installed.
"""
from gspread.utils import fill_gaps
from gspread import Cell
import pandas as pd
from pandas.io.parsers import TextParser
import logging
import re
from numbers import Real
from six import string_types, ensure_text

try:
    from collections.abc import defaultdict
except ImportError:
    from collections import defaultdict
try:
    from itertools import chain, zip_longest
except ImportError:
    from itertools import chain, izip_longest as zip_longest

logger = logging.getLogger(__name__)

__all__ = ("set_with_dataframe", "get_as_dataframe")

WORKSHEET_MAX_CELL_COUNT = 5000000

def _escaped_string(value, string_escaping):
    if value in (None, ""):
        return ""
    if string_escaping == "default":
        if value.startswith("'"):
            return "'%s" % value
    elif string_escaping == "off":
        return value
    elif string_escaping == "full":
        return "'%s" % value
    elif callable(string_escaping):
        if string_escaping(value):
            return "'%s" % value
    else:
        raise ValueError(
            "string_escaping parameter must be one of: "
            "'default', 'off', 'full', any callable taking one parameter"
        )
    return value


def _cellrepr(value, allow_formulas, string_escaping):
    """
    Get a string representation of dataframe value.

    :param :value: the value to represent
    :param :allow_formulas: if True, allow values starting with '='
            to be interpreted as formulas; otherwise, escape
            them with an apostrophe to avoid formula interpretation.
    """
    if pd.isnull(value) is True:
        return ""
    if isinstance(value, Real):
        return value
    if not isinstance(value, string_types):
        value = str(value)

    value = ensure_text(value, encoding='utf-8')

    if (not allow_formulas) and value.startswith("="):
        return "'%s" % value
    else:
        return _escaped_string(value, string_escaping)


def _resize_to_minimum(worksheet, rows=None, cols=None):
    """
    Resize the worksheet to guarantee a minimum size, either in rows,
    or columns, or both.

    Both rows and cols are optional.
    """
    current_rows, current_cols = (worksheet.row_count, worksheet.col_count)
    desired_rows, desired_cols = (rows, cols)
    if desired_rows is not None and desired_rows <= current_rows:
        desired_rows = current_rows
    if desired_cols is not None and desired_cols <= current_cols:
        desired_cols = current_cols
    resize_cols_first = False
    if desired_rows is not None and desired_cols is not None:
        # special case: if desired sheet size now > cell limit for sheet,
        # resize to exactly rows x cols, which in certain cases will 
        # allow worksheet to stay within cell limit.
        if desired_rows * desired_cols > WORKSHEET_MAX_CELL_COUNT:
            desired_rows, desired_cols = (rows, cols)

        # Large increase that requires exact re-sizing to avoid exceeding
        # cell limit might be, for example, 1000000 rows and 2 columns,
        # for a worksheet that currently has 100 rows and 26 columns..
        # The sheets API, however, applies new rowCount first, then
        # checks against cell count limit before applying new colCount!
        # In the above case, applying new rowCount produces 26 million
        # cells, the limit is exceeded, and API aborts the change and
        # returns a 400 response.
        # So to avoid a 400 response, we must in these cases have
        # _resize_to_minimum call resize twice, first with the value
        # that will reduce cell count and second with the value that
        # will increase cell count.
        # We don't seem to need to address the reversed case, where
        # columnCount is applied first, since Sheets API seems to apply
        # rowCount first in all cases. There is test coverage of this
        # reversed case, to guard against Sheets API changes in future.
        if (
            cols is not None and 
            cols < current_cols and 
            desired_rows * current_cols > WORKSHEET_MAX_CELL_COUNT
        ):
            resize_cols_first = True

    if desired_cols is not None or desired_rows is not None:
        if resize_cols_first:
            worksheet.resize(cols=desired_cols)
            worksheet.resize(rows=desired_rows)
        else:
            worksheet.resize(desired_rows, desired_cols)


def _get_all_values(worksheet, evaluate_formulas):
    data = worksheet.spreadsheet.values_get(
        worksheet.title,
        params={
            "valueRenderOption": (
                "UNFORMATTED_VALUE" if evaluate_formulas else "FORMULA"
            ),
            "dateTimeRenderOption": "FORMATTED_STRING",
        },
    )
    (row_offset, column_offset) = (1, 1)
    (last_row, last_column) = (worksheet.row_count, worksheet.col_count)
    values = data.get("values", [])

    rect_values = fill_gaps(
        values,
        rows=last_row - row_offset + 1,
        cols=last_column - column_offset + 1,
    )

    cells = [
        Cell(row=i + row_offset, col=j + column_offset, value=value)
        for i, row in enumerate(rect_values)
        for j, value in enumerate(row)
    ]

    # defaultdicts fill in gaps for empty rows/cells not returned by gdocs
    rows = defaultdict(lambda: defaultdict(str))
    for cell in cells:
        row = rows.setdefault(int(cell.row), defaultdict(str))
        row[cell.col] = cell.value

    if not rows:
        return []

    all_row_keys = chain.from_iterable(row.keys() for row in rows.values())
    rect_cols = range(1, max(all_row_keys) + 1)
    rect_rows = range(1, max(rows.keys()) + 1)

    return [[rows[i][j] for j in rect_cols] for i in rect_rows]


def get_as_dataframe(worksheet, evaluate_formulas=False, **options):
    r"""
    Returns the worksheet contents as a DataFrame.

    :param worksheet: the worksheet.
    :param evaluate_formulas: if True, get the value of a cell after
            formula evaluation; otherwise get the formula itself if present.
            Defaults to False.
    :param \*\*options: all the options for pandas.io.parsers.TextParser,
            according to the version of pandas that is installed.
            (Note: TextParser supports only the default 'python' parser engine,
            not the C engine.)
    :returns: pandas.DataFrame
    """
    all_values = _get_all_values(worksheet, evaluate_formulas)
    return TextParser(all_values, **options).read(options.get("nrows", None))


def _determine_level_count(index):
    if hasattr(index, "levshape"):
        return len(index.levshape)
    return 1

def _index_names(index):
    names = []
    if hasattr(index, "names"):
        names = [ i if i != None else "" for i in index.names ]
    elif index.name not in (None, ""):
        names = [index.name]
    if not any([n not in (None, "") for n in names]):
        names = []
    return names

def set_with_dataframe(
    worksheet,
    dataframe,
    row=1,
    col=1,
    include_index=False,
    include_column_header=True,
    resize=False,
    allow_formulas=True,
    string_escaping="default",
):
    """
    Sets the values of a given DataFrame, anchoring its upper-left corner
    at (row, col). (Default is row 1, column 1.)

    :param worksheet: the gspread worksheet to set with content of DataFrame.
    :param dataframe: the DataFrame.
    :param row: Row at which to start writing the DataFrame. Default is 1.
    :param col: Column  at which to start writing the DataFrame. Default is 1.
    :param include_index: if True, include the DataFrame's index as an
            additional column. Defaults to False.
    :param include_column_header: if True, add a header row or rows before data
            with column names. (If include_index is True, the index's name(s)
            will be used as its columns' headers.) Defaults to True.
    :param resize: if True, changes the worksheet's size to match the shape
            of the provided DataFrame. If False, worksheet will only be
            resized as necessary to contain the DataFrame contents.
            Defaults to False.
    :param allow_formulas: if True, interprets `=foo` as a formula in
            cell values; otherwise all text beginning with `=` is escaped
            to avoid its interpretation as a formula. Defaults to True.
    :param string_escaping: determines when string values are escaped as text
            literals (by adding an initial `'` character) in requests to
            Sheets API.
            Four parameter values are accepted:
              - 'default': only escape strings starting with a literal `'`
                           character
              - 'off': escape nothing; cell values starting with a `'` will be
                       interpreted by sheets as an escape character followed by
                       a text literal.
              - 'full': escape all string values
              - any callable object: will be called once for each cell's string
                     value; if return value is true, string will be escaped
                     with preceding `'` (A useful technique is to pass a
                     regular expression bound method, e.g.
                     `re.compile(r'^my_regex_.*$').search`.)
            The escaping done when allow_formulas=False (escaping string values
            beginning with `=`) is unaffected by this parameter's value.
            Default value is `'default'`.
    """
    # x_pos, y_pos refers to the position of data rows only,
    # excluding any header rows in the google sheet.
    # If header-related params are True, the values are adjusted
    # to allow space for the headers.
    y, x = dataframe.shape
    index_col_size = 0
    column_header_size = 0
    index_names = _index_names(dataframe.index)
    column_names_not_labels = _index_names(dataframe.columns)
    if include_index:
        index_col_size = _determine_level_count(dataframe.index)
        x += index_col_size
    if include_column_header:
        column_header_size = _determine_level_count(dataframe.columns)
        y += column_header_size
        # if included index has name(s) it needs its own header row to accommodate columns' index names
        if column_header_size > 1 and include_index and index_names:
            y += 1
    if resize:
        worksheet.resize(y, x)
    else:
        _resize_to_minimum(worksheet, y, x)

    updates = []

    if include_column_header:
        elts = list(dataframe.columns)
        # if columns object is multi-index, it will span multiple rows
        extra_header_row = None
        if column_header_size > 1:
            elts = list(dataframe.columns)
            if include_index:
                extra = tuple(column_names_not_labels) \
                        if column_names_not_labels \
                        else ("",) * column_header_size
                extra = [ extra ]
                if index_col_size > 1:
                    extra = extra + [ ("",) * column_header_size ] * (index_col_size - 1)
                elts = extra + elts
                # if index has names, they need their own header row
                if index_names:
                    extra_header_row = list(index_names) + [ "" ] * len(dataframe.columns)
            for level in range(0, column_header_size):
                for idx, tup in enumerate(elts):
                    updates.append(
                        (
                            row,
                            col + idx,
                            _cellrepr(
                                tup[level], allow_formulas, string_escaping
                            ),
                        )
                    )
                row += 1
            if extra_header_row:
                for idx, val in enumerate(extra_header_row):
                    updates.append(
                        (
                            row,
                            col + idx,
                            _cellrepr(
                                val, allow_formulas, string_escaping
                            ),
                        )
                    )
                row += 1

        else:
            # columns object is not multi-index, columns object's "names"
            # can not be written anywhere in header and be parseable to pandas.
            elts = list(dataframe.columns)
            if include_index:
                # if index has names, they do NOT need their own header row
                if index_names:
                    elts = index_names + elts
                else:
                    elts = ([""] * index_col_size) + elts
            for idx, val in enumerate(elts):
                updates.append(
                    (
                        row,
                        col + idx,
                        _cellrepr(val, allow_formulas, string_escaping),
                    )
                )
            row += 1

    values = []
    for value_row, index_value in zip_longest(
        dataframe.to_numpy('object'), dataframe.index.to_numpy('object')
    ):
        if include_index:
            if not isinstance(index_value, (list, tuple)):
                index_value = [index_value]
            value_row = list(index_value) + list(value_row)
        values.append(value_row)
    for y_idx, value_row in enumerate(values):
        for x_idx, cell_value in enumerate(value_row):
            updates.append(
                (
                    y_idx + row,
                    x_idx + col,
                    _cellrepr(cell_value, allow_formulas, string_escaping),
                )
            )

    if not updates:
        logger.debug("No updates to perform on worksheet.")
        return

    cells_to_update = [Cell(row, col, value) for row, col, value in updates]
    logger.debug("%d cell updates to send", len(cells_to_update))

    resp = worksheet.update_cells(
        cells_to_update, value_input_option="USER_ENTERED"
    )
    logger.debug("Cell update response: %s", resp)



# -*- coding: utf-8 -*-
from .mock_worksheet import (
    MockWorksheet,
    CELL_LIST,
    CELL_LIST_STRINGIFIED,
    CELL_LIST_STRINGIFIED_NO_THINGY,
)

from gspread_dataframe import get_as_dataframe, set_with_dataframe
from gspread_dataframe import _escaped_string as escape, _cellrepr as cellrepr
from gspread import Cell
import numpy as np
import pandas as pd
from difflib import SequenceMatcher

import unittest

try:
    from unittest.mock import Mock, MagicMock
except ImportError:
    from mock import Mock, MagicMock
from datetime import datetime
import re

# Expected results

COLUMN_NAMES = [
    "Thingy",
    "Syntax",
    "Numeric Column",
    "Formula Column",
    "Date Column",
    "Values are...",
    "Selection",
    "Label(s) referencible in chart title",
    "Dialect-specific implementations",
    "Notes",
]

USECOLS_COLUMN_NAMES = [
    "Thingy",
    "Numeric Column",
    "Formula Column",
    "Date Column",
]


# Tests


class TestStringEscaping(unittest.TestCase):
    CORE_VALUES = ("foo", '"""""', "2015-06-14", "345.60", "+", "=sum(a:a)")
    VALUES_WITH_LEADING_APOSTROPHE = ("'foo", "'")
    VALUES_NEVER_ESCAPED = ("",)

    def _run_values_for_escape_args(
        self, escape_arg, escaped_values, unescaped_values
    ):
        for value in escaped_values:
            self.assertEqual(escape(value, escape_arg), "'" + value)
        for value in unescaped_values:
            self.assertEqual(escape(value, escape_arg), value)
        for value in self.VALUES_NEVER_ESCAPED:
            self.assertEqual(escape(value, escape_arg), value)

    def test_default(self):
        self._run_values_for_escape_args(
            "default", self.VALUES_WITH_LEADING_APOSTROPHE, self.CORE_VALUES
        )

    def test_off(self):
        self._run_values_for_escape_args(
            "off", (), self.CORE_VALUES + self.VALUES_WITH_LEADING_APOSTROPHE
        )

    def test_full(self):
        self._run_values_for_escape_args(
            "full", self.CORE_VALUES + self.VALUES_WITH_LEADING_APOSTROPHE, ()
        )

    def test_callable(self):
        self._run_values_for_escape_args(
            lambda x: False,
            (),
            self.CORE_VALUES + self.VALUES_WITH_LEADING_APOSTROPHE,
        )
        self._run_values_for_escape_args(
            re.compile(r"@@@@@{200}").match,
            (),
            self.CORE_VALUES + self.VALUES_WITH_LEADING_APOSTROPHE,
        )
        self._run_values_for_escape_args(
            lambda x: True,
            self.CORE_VALUES + self.VALUES_WITH_LEADING_APOSTROPHE,
            (),
        )
        self._run_values_for_escape_args(
            re.compile(r".*").match,
            self.CORE_VALUES + self.VALUES_WITH_LEADING_APOSTROPHE,
            (),
        )

    def test_formula_cellrepr_when_no_formulas_allowed(self):
        self.assertEqual(cellrepr("=A1", allow_formulas=False, string_escaping="default"), "'=A1")


class TestWorksheetReads(unittest.TestCase):
    def setUp(self):
        self.sheet = MockWorksheet()

    def test_noargs(self):
        df = get_as_dataframe(self.sheet)
        self.assertEqual(list(df.columns.values), COLUMN_NAMES)
        self.assertEqual(len(df.columns), 10)
        self.assertEqual(len(df), 9)
        self.assertEqual(df.index.name, None)
        self.assertEqual(type(df.index).__name__, "RangeIndex")
        self.assertEqual(list(df.index.values), list(range(9)))

    def test_evaluate_formulas_true(self):
        df = get_as_dataframe(self.sheet, evaluate_formulas=True)
        self.assertEqual(list(df.columns.values), COLUMN_NAMES)
        self.assertEqual(df["Formula Column"][0], 2.226)

    def test_evaluate_formulas_false(self):
        df = get_as_dataframe(self.sheet)
        self.assertEqual(list(df.columns.values), COLUMN_NAMES)
        self.assertEqual(df["Formula Column"][0], "=R[0]C[-1]*2")

    def test_usecols(self):
        df = get_as_dataframe(self.sheet, usecols=USECOLS_COLUMN_NAMES)
        self.assertEqual(list(df.columns.values), USECOLS_COLUMN_NAMES)

    def test_indexcol(self):
        df = get_as_dataframe(self.sheet, index_col=4)
        self.assertEqual(len(df.columns), 9)
        self.assertEqual(df.index.name, "Date Column")
        self.assertEqual(type(df.index).__name__, "Index")
        self.assertEqual(df.index.values[0], "2017-03-04")

    def test_indexcol_none(self):
        df = get_as_dataframe(self.sheet, index_col=False)
        self.assertEqual(len(df.columns), 10)
        self.assertEqual(df.index.name, None)
        self.assertEqual(type(df.index).__name__, "RangeIndex")
        self.assertEqual(list(df.index.values), list(range(9)))

    def test_header_false(self):
        df = get_as_dataframe(self.sheet, header=None)
        self.assertEqual(len(df), 10)

    def test_header_first_row(self):
        df = get_as_dataframe(self.sheet, header=0)
        self.assertEqual(len(df), 9)

    def test_skiprows(self):
        df = get_as_dataframe(self.sheet, skiprows=range(1, 4))
        self.assertEqual(len(df), 6)

    def test_prefix(self):
        df = get_as_dataframe(
            self.sheet, skiprows=[0], header=None, prefix="COL"
        )
        self.assertEqual(len(df), 9)
        self.assertEqual(
            df.columns.tolist(), ["COL" + str(i) for i in range(10)]
        )

    def test_squeeze(self):
        df = get_as_dataframe(self.sheet, usecols=[0], squeeze=True)
        self.assertTrue(isinstance(df, pd.Series))
        self.assertEqual(len(df), 9)

    def test_converters_datetime(self):
        df = get_as_dataframe(
            self.sheet,
            converters={
                "Date Column": lambda x: datetime.strptime(x, "%Y-%m-%d")
            },
        )
        self.assertEqual(df["Date Column"][0], datetime(2017, 3, 4))

    def test_dtype_raises(self):
        self.assertRaises(
            ValueError,
            get_as_dataframe,
            self.sheet,
            dtype={"Numeric Column": np.float64},
        )

    def test_no_nafilter(self):
        df = get_as_dataframe(self.sheet, na_filter=False)
        self.assertEqual(df["Dialect-specific implementations"][7], "")

    def test_nafilter(self):
        df = get_as_dataframe(self.sheet, na_filter=True)
        self.assertTrue(np.isnan(df["Dialect-specific implementations"][7]))

    def test_parse_dates_true(self):
        df = get_as_dataframe(self.sheet, index_col=4, parse_dates=True)
        self.assertEqual(df.index[0], pd.Timestamp("2017-03-04 00:00:00"))

    def test_parse_dates_true_infer(self):
        df = get_as_dataframe(
            self.sheet,
            index_col=4,
            parse_dates=True,
            infer_datetime_format=True,
        )
        self.assertEqual(df.index[0], pd.Timestamp("2017-03-04 00:00:00"))

    def test_parse_dates_custom_parser(self):
        df = get_as_dataframe(
            self.sheet,
            parse_dates=[4],
            date_parser=lambda x: datetime.strptime(x, "%Y-%m-%d"),
        )
        self.assertEqual(df["Date Column"][0], datetime(2017, 3, 4))


_original_mock_failure_message = Mock._format_mock_failure_message


def _format_mock_failure_message(self, args, kwargs):
    message = "Expected call: %s\nActual call: %s"
    expected_string = self._format_mock_call_signature(args, kwargs)
    call_args = self.call_args
    if len(call_args) == 3:
        call_args = call_args[1:]
    actual_string = self._format_mock_call_signature(*call_args)
    msg = message % (expected_string, actual_string)
    if (
        len(call_args[0]) > 1
        and isinstance(call_args[0][1], (str, bytes))
        and len(args) > 1
        and isinstance(args[1], (str, bytes))
        and call_args[0][1] != args[1]
    ):
        import difflib

        sm = difflib.SequenceMatcher(None, call_args[0][1], args[1])
        m = sm.find_longest_match(0, len(call_args[0][1]), 0, len(args[1]))
        msg += "; diff: at index %d, expected %s -- actual %s" % (
            m.a + m.size,
            call_args[0][1][m.a + m.size - 40 : m.a + m.size + 40],
            args[1][m.b + m.size - 40 : m.b + m.size + 40],
        )
    return msg


Mock._format_mock_failure_message = _format_mock_failure_message

# have to patch Cell to make cells comparable
def __eq__(self, other):
    if not isinstance(other, self.__class__):
        return False
    return (
        self.row == other.row
        and self.col == other.col
        and self.value == other.value
    )


Cell.__eq__ = __eq__


class TestWorksheetWrites(unittest.TestCase):
    def setUp(self):
        self.sheet = MockWorksheet()
        self.sheet.resize = MagicMock()
        self.sheet.update_cells = MagicMock()
        self.sheet.spreadsheet.values_update = MagicMock()

    def test_write_basic(self):
        df = get_as_dataframe(self.sheet, na_filter=False)
        set_with_dataframe(
            self.sheet,
            df,
            resize=True,
            string_escaping=re.compile(r"3e50").match,
        )
        self.sheet.resize.assert_called_once_with(10, 10)
        self.sheet.update_cells.assert_called_once_with(
            CELL_LIST_STRINGIFIED, value_input_option="USER_ENTERED"
        )

    def test_include_index_false(self):
        df = get_as_dataframe(self.sheet, na_filter=False)
        df_index = df.set_index("Thingy")
        set_with_dataframe(
            self.sheet,
            df_index,
            resize=True,
            include_index=False,
            string_escaping=lambda x: x == "3e50",
        )
        self.sheet.resize.assert_called_once_with(10, 9)
        self.sheet.update_cells.assert_called_once_with(
            CELL_LIST_STRINGIFIED_NO_THINGY, value_input_option="USER_ENTERED"
        )

    def test_include_index_true(self):
        df = get_as_dataframe(self.sheet, na_filter=False)
        df_index = df.set_index("Thingy")
        set_with_dataframe(
            self.sheet,
            df_index,
            resize=True,
            include_index=True,
            string_escaping=re.compile(r"3e50").match,
        )
        self.sheet.resize.assert_called_once_with(10, 10)
        self.sheet.update_cells.assert_called_once_with(
            CELL_LIST_STRINGIFIED, value_input_option="USER_ENTERED"
        )

    def test_write_list_value_to_cell(self):
        df = get_as_dataframe(self.sheet, na_filter=False)
        df.at[0, "Numeric Column"] = [1, 2, 3]
        set_with_dataframe(
            self.sheet,
            df,
            resize=True,
            string_escaping=re.compile(r"3e50").match,
        )
        self.sheet.resize.assert_called_once_with(10, 10)
        self.sheet.update_cells.assert_called_once_with(
            CELL_LIST_STRINGIFIED, value_input_option="USER_ENTERED"
        )



import os.path
import json
import re
from gspread import Cell
from gspread_dataframe import _cellrepr


def contents_of_file(filename, et_parse=True):
    with open(os.path.join(os.path.dirname(__file__), filename), "r") as f:
        return json.load(f)


SHEET_CONTENTS_FORMULAS = contents_of_file("sheet_contents_formulas.json")
SHEET_CONTENTS_EVALUATED = contents_of_file("sheet_contents_evaluated.json")
CELL_LIST = [
    Cell(row=i + 1, col=j + 1, value=value)
    for i, row in enumerate(contents_of_file("cell_list.json"))
    for j, value in enumerate(row)
]

CELL_LIST_STRINGIFIED = [
    Cell(
        row=i + 1,
        col=j + 1,
        value=_cellrepr(
            value,
            allow_formulas=True,
            string_escaping=re.compile(r"3e50").match,
        ),
    )
    for i, row in enumerate(contents_of_file("cell_list.json"))
    for j, value in enumerate(row)
]

_without_index = contents_of_file("cell_list.json")
for _r in _without_index:
    del _r[0]

CELL_LIST_STRINGIFIED_NO_THINGY = [
    Cell(
        row=i + 1,
        col=j + 1,
        value=_cellrepr(
            value,
            allow_formulas=True,
            string_escaping=re.compile(r"3e50").match,
        ),
    )
    for i, row in enumerate(_without_index)
    for j, value in enumerate(row)
]


class MockWorksheet(object):
    def __init__(self):
        self.row_count = 10
        self.col_count = 10
        self.id = "fooby"
        self.title = "gspread dataframe test"
        self.spreadsheet = MockSpreadsheet()


class MockSpreadsheet(object):
    def values_get(self, *args, **kwargs):
        if (
            kwargs.get("params", {}).get("valueRenderOption")
            == "UNFORMATTED_VALUE"
        ):
            return SHEET_CONTENTS_EVALUATED
        if kwargs.get("params", {}).get("valueRenderOption") == "FORMULA":
            return SHEET_CONTENTS_FORMULAS


if __name__ == "__main__":
    from gspread_dataframe import *

    ws = MockWorksheet()



# -*- coding: utf-8 -*-

import os
import re
import random
import unittest
import itertools
import uuid
import json
import logging
import sys
from random import uniform
from datetime import datetime, date
from gspread.exceptions import APIError
import pandas as pd
from gspread_dataframe import \
    get_as_dataframe, \
    set_with_dataframe, \
    _resize_to_minimum

try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser

from oauth2client.service_account import ServiceAccountCredentials

import gspread
from gspread import utils

try:
    unicode
except NameError:
    basestring = unicode = str

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

CONFIG_FILENAME = os.path.join(os.path.dirname(__file__), "tests.config")
CREDS_FILENAME = os.path.join(os.path.dirname(__file__), "creds.json")
SCOPE = [
    "https://spreadsheets.google.com/feeds",
    "https://www.googleapis.com/auth/drive.file",
]

I18N_STR = u"Iñtërnâtiônàlizætiøn"  # .encode('utf8')

CELL_LIST_FILENAME = os.path.join(os.path.dirname(__file__), "cell_list.json")

STRING_ESCAPING_PATTERN = re.compile(r"(?:'\+|3e50)").match


def read_config(filename):
    config = ConfigParser.ConfigParser()
    with open(filename) as fp:
        if hasattr(config, "read_file"):
            read_func = config.read_file
        else:
            read_func = config.readfp
        read_func(fp)
    return config


def read_credentials(filename):
    return ServiceAccountCredentials.from_json_keyfile_name(filename, SCOPE)


def gen_value(prefix=None):
    if prefix:
        return u"%s %s" % (prefix, gen_value())
    else:
        return unicode(uuid.uuid4())


class GspreadDataframeTest(unittest.TestCase):
    config = None
    gc = None

    @classmethod
    def setUpClass(cls):
        try:
            cls.config = read_config(CONFIG_FILENAME)
            credentials = read_credentials(CREDS_FILENAME)
            cls.gc = gspread.authorize(credentials)
        except IOError as e:
            msg = "Can't find %s for reading test configuration. "
            raise Exception(msg % e.filename)

    def setUp(self):
        if self.__class__.gc is None:
            self.__class__.setUpClass()
        self.assertTrue(isinstance(self.gc, gspread.client.Client))


class WorksheetTest(GspreadDataframeTest):
    """Test for gspread_dataframe using a gspread.Worksheet."""

    spreadsheet = None

    @classmethod
    def setUpClass(cls):
        super(WorksheetTest, cls).setUpClass()
        ss_id = cls.config.get("Spreadsheet", "id")
        cls.spreadsheet = cls.gc.open_by_key(ss_id)
        cls.spreadsheet.batch_update(
            {
                "requests": [
                    {
                        "updateSpreadsheetProperties": {
                            "properties": {"locale": "en_US"},
                            "fields": "locale",
                        }
                    }
                ]
            }
        )
        try:
            test_sheet = cls.spreadsheet.worksheet("wksht_int_test")
            if test_sheet:
                # somehow left over from interrupted test, remove.
                cls.spreadsheet.del_worksheet(test_sheet)
        except gspread.exceptions.WorksheetNotFound:
            pass  # expected

    def setUp(self):
        super(WorksheetTest, self).setUp()
        self.streamHandler = logger.addHandler(logging.StreamHandler(sys.stdout))
        if self.__class__.spreadsheet is None:
            self.__class__.setUpClass()
        self.sheet = self.spreadsheet.add_worksheet("wksht_int_test", 20, 20)
        self.__class__.spreadsheet.batch_update(
            {
                "requests": [
                    {
                        "updateSpreadsheetProperties": {
                            "properties": {"locale": "en_US"},
                            "fields": "locale",
                        }
                    }
                ]
            }
        )

    def tearDown(self):
        self.spreadsheet.del_worksheet(self.sheet)
        logger.removeHandler(self.streamHandler)

    def test_roundtrip(self):
        # populate sheet with cell list values
        rows = None
        with open(CELL_LIST_FILENAME) as f:
            rows = json.load(f)

        self.sheet.resize(10, 10)
        cell_list = self.sheet.range("A1:J10")
        for cell, value in zip(cell_list, itertools.chain(*rows)):
            cell.value = value
        self.sheet.update_cells(cell_list)

        df = get_as_dataframe(self.sheet)
        set_with_dataframe(
            self.sheet, df, string_escaping=STRING_ESCAPING_PATTERN
        )
        df2 = get_as_dataframe(self.sheet)
        self.assertTrue(df.equals(df2))

    def test_numeric_values_with_spanish_locale(self):
        # set locale!
        self.__class__.spreadsheet.batch_update(
            {
                "requests": [
                    {
                        "updateSpreadsheetProperties": {
                            "properties": {"locale": "es_ES"},
                            "fields": "locale",
                        }
                    }
                ]
            }
        )
        # populate sheet with cell list values
        rows = None
        with open(CELL_LIST_FILENAME) as f:
            rows = json.load(f)

        self.sheet.resize(10, 10)
        cell_list = self.sheet.range("A1:J10")
        for cell, value in zip(cell_list, itertools.chain(*rows)):
            cell.value = value
        self.sheet.update_cells(cell_list)

        df = get_as_dataframe(self.sheet)
        set_with_dataframe(
            self.sheet, df, string_escaping=STRING_ESCAPING_PATTERN
        )
        df2 = get_as_dataframe(self.sheet)
        # check that some numeric values in numeric column are intact
        self.assertEqual(3.804, df2["Numeric Column"][3])
        self.assertTrue(df.equals(df2))

    def test_nrows(self):
        # populate sheet with cell list values
        rows = None
        with open(CELL_LIST_FILENAME) as f:
            rows = json.load(f)

        self.sheet.resize(10, 10)
        cell_list = self.sheet.range("A1:J10")
        for cell, value in zip(cell_list, itertools.chain(*rows)):
            cell.value = value
        self.sheet.update_cells(cell_list)

        for nrows in (9, 6, 0):
            df = get_as_dataframe(self.sheet, nrows=nrows)
            self.assertEqual(nrows, len(df))


    def test_resize_to_minimum_large(self):
        self.sheet.resize(100, 26)
        self.sheet = self.sheet.spreadsheet.worksheet(self.sheet.title)
        # Large increase that requires exact re-sizing to avoid exceeding 
        # cell limit: this should result in 1000000 rows and 2 columns.
        # The sheets API, however, applies new rowCount first, then
        # checks against cell count limit before applying new colCount!
        # So to avoid a 400 response, we must in these cases have
        # _resize_to_minimum call resize twice, first with the value
        # that will reduce cell count and second with the value that
        # will increase cell count.
        _resize_to_minimum(self.sheet, 1000000, 2)
        self.sheet = self.sheet.spreadsheet.worksheet(self.sheet.title)
        self.assertEqual(1000000, self.sheet.row_count)
        self.assertEqual(2, self.sheet.col_count)
        # let's test the other case, where if columnCount were applied
        # first the limit would be exceeded.
        _resize_to_minimum(self.sheet, 10000, 26)
        self.sheet = self.sheet.spreadsheet.worksheet(self.sheet.title)
        self.assertEqual(10000, self.sheet.row_count)
        self.assertEqual(26, self.sheet.col_count)

    def test_resize_to_minimum(self):
        self.sheet.resize(100, 26)
        self.sheet = self.sheet.spreadsheet.worksheet(self.sheet.title)
        # min rows < current, no change
        _resize_to_minimum(self.sheet, 20, None)
        self.sheet = self.sheet.spreadsheet.worksheet(self.sheet.title)
        self.assertEqual(100, self.sheet.row_count)
        self.assertEqual(26, self.sheet.col_count)
        # min cols < current, no change
        _resize_to_minimum(self.sheet, None, 2)
        self.sheet = self.sheet.spreadsheet.worksheet(self.sheet.title)
        self.assertEqual(100, self.sheet.row_count)
        self.assertEqual(26, self.sheet.col_count)
        # increase rows
        _resize_to_minimum(self.sheet, 200, None)
        self.sheet = self.sheet.spreadsheet.worksheet(self.sheet.title)
        self.assertEqual(200, self.sheet.row_count)
        self.assertEqual(26, self.sheet.col_count)
        # increase cols
        _resize_to_minimum(self.sheet, None, 27)
        self.sheet = self.sheet.spreadsheet.worksheet(self.sheet.title)
        self.assertEqual(200, self.sheet.row_count)
        self.assertEqual(27, self.sheet.col_count)
        # increase both
        _resize_to_minimum(self.sheet, 201, 28)
        self.sheet = self.sheet.spreadsheet.worksheet(self.sheet.title)
        self.assertEqual(201, self.sheet.row_count)
        self.assertEqual(28, self.sheet.col_count)
        # large increase that exact re-sizing cannot keep below cell limit
        # this should result in a 400 ApiError
        with self.assertRaises(APIError):
            _resize_to_minimum(self.sheet, 1000000, None)
        
    def test_multiindex(self):
        # populate sheet with cell list values
        rows = None
        with open(CELL_LIST_FILENAME) as f:
            rows = json.load(f)
        mi = list(
            pd.MultiIndex.from_product(
                [["A", "B"], ["one", "two", "three", "four", "five"]]
            )
        )
        column_names = ["Category", "Subcategory"] + rows[0]
        rows = [column_names] + [
            list(index_tup) + row for row, index_tup in zip(rows[1:], mi)
        ]
        cell_list = self.sheet.range("A1:L10")
        for cell, value in zip(cell_list, itertools.chain(*rows)):
            cell.value = value
        self.sheet.update_cells(cell_list)
        self.sheet.resize(10, 12)
        self.sheet = self.sheet.spreadsheet.worksheet(self.sheet.title)
        df = get_as_dataframe(self.sheet, index_col=[0, 1])
        set_with_dataframe(
            self.sheet,
            df,
            resize=True,
            include_index=True,
            string_escaping=STRING_ESCAPING_PATTERN,
        )
        # must do this to refresh the size attributes of worksheet
        self.sheet = self.sheet.spreadsheet.worksheet(self.sheet.title)
        df2 = get_as_dataframe(self.sheet, index_col=[0, 1])
        self.assertTrue(df.equals(df2))

    def test_multiindex_column_header(self):
        # populate sheet with cell list values
        rows = None
        with open(CELL_LIST_FILENAME) as f:
            rows = json.load(f)
        column_headers = [
            "SQL",
            "SQL",
            "SQL",
            "SQL",
            "SQL",
            "Misc",
            "Misc",
            "Misc",
            "Misc",
            "Misc",
        ]
        rows = [column_headers] + rows
        cell_list = self.sheet.range("A1:J11")
        for cell, value in zip(cell_list, itertools.chain(*rows)):
            cell.value = value
        self.sheet.update_cells(cell_list)
        self.sheet.resize(11, 10)
        self.sheet = self.sheet.spreadsheet.worksheet(self.sheet.title)
        df = get_as_dataframe(self.sheet, header=[0, 1])
        self.assertEqual((2, 10), getattr(df.columns, "levshape", None)),
        set_with_dataframe(
            self.sheet,
            df,
            resize=True,
            string_escaping=STRING_ESCAPING_PATTERN,
        )
        df2 = get_as_dataframe(self.sheet, header=[0, 1])
        self.assertTrue(df.equals(df2))

    def test_int64_json_issue35(self):
        df = pd.DataFrame(
            {
                'a':pd.Series([1, 2, 3],dtype='int64',index=pd.RangeIndex(start=0, stop=3, step=1)),
                'b':pd.Series([4, 5, 6],dtype='int64',index=pd.RangeIndex(start=0, stop=3, step=1))
            },
            index=pd.RangeIndex(start=0, stop=3, step=1)
        )
        set_with_dataframe(
            self.sheet,
            df,
            resize=True,
            include_index=True
        )
        self.sheet = self.sheet.spreadsheet.worksheet(self.sheet.title)
        df2 = get_as_dataframe(self.sheet, dtype={'a': 'int64', 'b': 'int64'}, index_col=0, header=0)
        self.assertTrue(df.equals(df2))

    def test_header_writing_and_parsing(self):
        truth_table = itertools.product(*([[False, True]] * 4))
        for include_index, columns_multilevel, index_has_names, columns_has_names in truth_table:
            logger.info(
                "Testing include_index %s, index_has_names %s, columns_multilevel %s, columns_has_names %s",
                include_index, index_has_names, columns_multilevel, columns_has_names
            )
            data = [[uniform(0, 100000) for i in range(8)] for j in range(20)]
            index_names = ["Category", "Subcategory"] if index_has_names else None
            index = list(
                itertools.product(
                    ["A", "B", "C", "D"],
                    ["one", "two", "three", "four", "five"],
                )
            )
            index = pd.MultiIndex.from_tuples(index, names=index_names)
            if not include_index:
                index = None
            columns = ["Alice", "Bob", "Carol", "Dave", "Ellen", "Fulgencio", "Gina", "Hector"]
            columns = [ ("Helpful" if i < 4 else "Unhelpful", v) for i, v in enumerate(columns) ]
            names = ["Demeanor", "Name"] if columns_has_names else None
            columns = pd.MultiIndex.from_tuples(columns, names=names)
            if not columns_multilevel:
                columns = columns.droplevel(0)
            df = pd.DataFrame.from_records(data, index=index, columns=columns)
            set_with_dataframe(self.sheet, df, resize=True, include_index=include_index)
            self.sheet = self.sheet.spreadsheet.worksheet(self.sheet.title)
            header_arg = list(range(len(getattr(columns, "levshape", [1]))))
            # if include_index and columns_multilevel and index_has_names, there
            # will be an additional header row
            index_col_arg = list(range(len(getattr(index, "levshape", [1]))))
            logger.info("header=%s, index_col=%s", header_arg, index_col_arg)
            df_readback = get_as_dataframe(
                self.sheet, 
                header=header_arg,
                index_col=(index_col_arg if include_index else None)
            )
            logger.info("DataFrames equal: %s", df.equals(df_readback))
            if not df.equals(df_readback):
                logger.info("%s", df)
                logger.info("%s", df.dtypes)
                logger.info("%s", df_readback)
                logger.info("%s", df_readback.dtypes)
            self.assertTrue(df.equals(df_readback))

